import unittest
from detektor.languageparser.java import JavaLanguageParser


class TestJavaLanguageParser(unittest.TestCase):

    def test_keywords_sanity(self):
        parser = JavaLanguageParser()

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'if(true)')
        self.assertEquals(parseresult.keywords['if'], 1)
        for keyword in parser.keywords:
            if keyword != 'if':
                self.assertEquals(parseresult.keywords[keyword], 0)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'if(x == 10) return')
        self.assertEquals(parseresult.keywords['if'], 1)
        self.assertEquals(parseresult.keywords['return'], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'while(true) return')
        self.assertEquals(parseresult.keywords['while'], 1)
        self.assertEquals(parseresult.keywords['return'], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'for(int x: somelist) return')
        self.assertEquals(parseresult.keywords['for'], 1)
        self.assertEquals(parseresult.keywords['int'], 1)
        self.assertEquals(parseresult.keywords['return'], 1)

    def test_operators_sanity(self):
        parser = JavaLanguageParser()

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a == 10')
        self.assertEquals(parseresult.operators['=='], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a==10')
        self.assertEquals(parseresult.operators['=='], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a*10')
        self.assertEquals(parseresult.operators['*'], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a ==10')
        self.assertEquals(parseresult.operators['=='], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a== 10')
        self.assertEquals(parseresult.operators['=='], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a > 10')
        self.assertEquals(parseresult.operators['>'], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a != 10')
        self.assertEquals(parseresult.operators['!='], 1)

    def test_parsed_functions(self):
        # Note: Only a sanity test. This only needs to ensure that
        # the function extracter is configured correctly, and that
        # the parser is exectuted on the extracted functions. The function
        # extracter and parser has their own tests.
        parser = JavaLanguageParser()

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, '''
        public class Test {
            public void main(String[] args) {
                printHello();
            }

            private void printHello() {
                System.out.println("Hello World!");
            }

            void test(int i) {
                if(i > 10) {
                    return 1;
                } else {
                    return 2;
                }
            }
        }
        ''')
        self.assertEquals(len(parseresult.parsed_functions), 3)
        self.assertEquals(parseresult.parsed_functions[0].label, 'void main(String[] args)')
        self.assertEquals(parseresult.parsed_functions[1].label, 'void printHello()')
        self.assertEquals(parseresult.parsed_functions[2].label, 'void test(int i)')
        self.assertEquals(parseresult.parsed_functions[2].keywords['if'], 1)
        self.assertEquals(parseresult.parsed_functions[2].keywords['else'], 1)
        self.assertEquals(parseresult.parsed_functions[2].keywords['return'], 2)




class ExtractedFunction(object):
    def __init__(self, name, sourcecode):
        """
        Parameters:
            name (unicode): The name of the function.
             Should contain as much unique info as possible about
             the function. This means that for languages where
             a function with the same name but different arguments
             or return types denotes separate functions, the arguments and return
             type should be included in the name. Should not contain any newlines,
             and all whitespace should be collapsed to a single space.
            description (unicode): The sourcecode of the function.
              This should not include the signature, only the body of the
              function.
        """
        self.name = name
        self.sourcecode = sourcecode

    def __unicode__(self):
        return u'{} {}'.format(self.name, self.sourcecode)

    def __str__(self):
        return unicode(self).encode('ascii', 'replace')

    def __repr__(self):
        return 'ExtractedFunction({})'.format(self)


class BaseFunctionExtractor(object):
    """
    Abstract base class for function extractors.

    A function extractor takes a sourcecode blob and
    extracts a list of functions and methods from it.
    """
    def extract(self, sourcecode):
        """
        Extract the functions in ``sourcecode`` and return the result as a
        list of :class:`ExtractedFunction` objects.

        Must be overridden in subclasses.
        """
        raise NotImplementedError()



import re
from detektor.functionextractor.base import BaseFunctionExtractor
from detektor.functionextractor.base import ExtractedFunction


class ClikeFunctionExtractor(BaseFunctionExtractor):
    functionstartpattern = re.compile(
        r'(?P<returntype>\w+[^ ]*)?\s+(?P<name>\w+)\w*\((?P<arguments>.*?)\)\s*\{')

    def __init__(self, keywords):
        self.keywords = keywords

    def _find_function_end(self, sourcecode, startindex):
        blocklevel = 1
        for index in xrange(startindex, len(sourcecode)):
            character = sourcecode[index]
            if character == '{':
                blocklevel += 1
            elif character == '}':
                blocklevel -= 1
            if blocklevel == 0:
                return index
        return len(sourcecode) - 1

    def extract(self, sourcecode):
        functions = []
        for match in self.functionstartpattern.finditer(sourcecode):
            groupdict = match.groupdict()
            if groupdict['name'] not in self.keywords:
                function_startindex = match.end()
                function_endindex = self._find_function_end(sourcecode, function_startindex)
                function_sourcecode = sourcecode[function_startindex:function_endindex]
                functions.append(ExtractedFunction(
                    name=u'{returntype} {name}({arguments})'.format(**match.groupdict()),
                    sourcecode=function_sourcecode
                ))
        return functions



import re
from detektor.codefilter.base import CodeFilterBase


class NormalizeOperatorWhitespace(CodeFilterBase):
    """
    Filter to make put a single space between all operators
    and their surrounding code.

    Makes simple lexers like shlex treat the opeators as words.

    Requires the language parser object to have ``operators`` and
    ``keywords`` attributes (see :class:`detektor.languageparser.shlexparser.ShlexLanguageParserBase`)
    """
    def __init__(self, languageparser):
        super(NormalizeOperatorWhitespace, self).__init__(languageparser)
        normalize_whitespace_operators = [re.escape(operator) \
                                          for operator in languageparser.operators]
        self.normalize_whitespace_pattern = re.compile(
            r'(\w)\s*(' +
            '|'.join(normalize_whitespace_operators) +
            r')\s*(\w)')

    def filter(self, sourcecode):
        return self.normalize_whitespace_pattern.sub(r'\1 \2 \3', sourcecode)



import shlex

from detektor.languageparser.base import LanguageParserBase
from detektor.parseresult import EditableParseResult


class ShlexLanguageParserBase(LanguageParserBase):
    #: Set of operators supported by the language.
    #: # E.g.: ``operators = {'<', '==', '>', '&&'}``
    operators = None

    #: Set of keywords supported by the language.
    #: # E.g.: ``keywords = {'if', 'for', 'and', 'or'}``
    keywords = None

    def __init__(self):
        super(ShlexLanguageParserBase, self).__init__()

    def get_parseresultclass(self):
        return EditableParseResult

    def make_parseresult(self, codeblocktype='program', label=None):
        parseresult_class = self.get_parseresultclass()
        return parseresult_class(
            set_of_all_operators=self.operators,
            set_of_all_keywords=self.keywords,
            codeblocktype=codeblocktype,
            label=label)

    def get_shlexobject(self, sourcecode):
        """
        Get a ``shlex.shlex`` object. Can be overridden to
        customize the lexer.
        """
        lexer = shlex.shlex(sourcecode)
        lexer.wordchars += ''.join(self.operators)
        return lexer

    def shlex_parse_token(self, parseresult, token):
        if token in self.keywords:
            parseresult.add_keyword(token)
        elif token in self.operators:
            parseresult.add_operator(token)

    def parse_program(self, parseresult, sourcecode):
        lexer = self.get_shlexobject(sourcecode)
        while True:
            token = lexer.get_token()
            if not token:
                break
            self.shlex_parse_token(parseresult, token)



from detektor.codefilter.normalizeoperatorwhitespace import NormalizeOperatorWhitespace
from detektor.languageparser.shlexparser import ShlexLanguageParserBase
import keyword


class PythonLanguageParser(ShlexLanguageParserBase):
    operators = {
        '!', '+', '-', '*', '**', '/', '//', '%', '<<', '>>', '&', '|',
        '=', '%=', '/=', '//=', '**=', '*=', '-=',
        '^', '~', '<', '>', '<=', '>=', '==', '!=', '<>',
    }
    keywords = set(keyword.kwlist)
    sourcecode_preprocessor_classes = [
        NormalizeOperatorWhitespace
    ]





class LanguageParserBase(object):
    """
    Abstract base class for parsing sourcecode.

    A parser is typically initialized once and used multiple times
    in the same thread.


    Example of how to use an implementation of LanguageParserBase::

        parser = SomeLanguageParser()
        sourcecodes = [
            'print "hello world"',
            'print "This is a test"'
        ]

        # Treat both the sourcecodes as a single "program",
        # and collect the results in a single EditableParseResult object:
        parseresult = parser.make_parseresult()
        for sourcecode in sourcecodes:
            parser.parse(sourcecode, parseresult)
        print parseresult

        # ... or treat them as separate programs:
        results = []
        for sourcecode in sourcecodes:
            parseresult = parser.make_parseresult()
            parser.parse(sourcecode, parseresult)
            results.append(parseresult)


    An important thing to notice here is that the parser is
    initialized once and used many times. Some parsers may do
    expensive optimizations when initialized, so it is a
    good idea to re-use the parser instance instead of recreating
    it each time.

    .. warning::

        A LanguageParserBase instance is not thread safe. It
        stores state on the object during each call to :meth:`.parse`.
    """

    #: List of preprocessors
    sourcecode_preprocessor_classes = []

    def __init__(self):
        """
        Parameters:
            sourcecode: The source code to parse.
            codeblocktype: Describes what the sourcecode is. Must be one of: "program", "function".
            label: A label that can be set to give the EditableParseResult a name/context.
                Typically used to give functions a name, but can also be used when
                ``codeblocktype`` is ``program``.
        """
        self.sourcecode_preprocessors = [cls(self) for cls in self.sourcecode_preprocessor_classes]

    def preprocess_sourcecode(self, sourcecode):
        """
        You can override this to preprocess the sourcecode before
        the parsing is started.
        """
        for sourcecode_preprocessor in self.sourcecode_preprocessors:
            sourcecode = sourcecode_preprocessor.filter(sourcecode)
        return sourcecode

    def make_parseresult(self, codeblocktype, label):
        """
        Parameters:
            codeblocktype: Describes what the sourcecode is. Must be one of: "program", "function".
            label: A label that can be set to give the EditableParseResult a name/context.
                Typically used to give functions a name, but can also be used when
                ``codeblocktype`` is ``program``.

        Returns:
            A :class:`detektor.parseresult.EditableParseResult` object.
        """
        raise NotImplementedError()

    def parse_program(self, sourcecode, parseresult):
        raise NotImplementedError()

    def get_function_sourcecode_parser_class(self):
        """
        Returns the class to use to parse functions.
        Can be overridden if you need to use a different class
        to parse functions than the one you use to parse the
        """
        return self.__class__

    def extract_functionsourcecode(self, sourcecode):
        """
        Extract the sourcecode of functions from the ``sourcecode`` and
        return a list of ``(functionname, functionsourcecode)`` tuples.
        """
        return []

    def parse_functions(self, sourcecode):
        function_sourcecode_parser_class = self.get_function_sourcecode_parser_class()
        parsed_functions = []
        for extractedfunction in self.extract_functionsourcecode(sourcecode):
            parseresult = self.make_parseresult(
                codeblocktype='function', label=extractedfunction.name)
            function_sourcecode_parser_class().parse(parseresult, extractedfunction.sourcecode)
            parsed_functions.append(parseresult)
        return parsed_functions

    def parse(self, parseresult, sourcecode):
        """
        Parse the ``sourcecode``, add the results to parseresult. Must be
        overridden in subclasses.

        Parameters:
            sourcecode: The sourcecode to parse as a string/unicode.
            parseresult: A :class:`detector.parser_result.EditableParseResult` object.
        """
        sourcecode = self.preprocess_sourcecode(sourcecode)
        self.parse_program(parseresult, sourcecode)
        if parseresult.is_program():
            parseresult.parsed_functions = self.parse_functions(sourcecode)



import unittest
from detektor.languageparser.python import PythonLanguageParser


class TestPythonLanguageParser(unittest.TestCase):

    def test_keywords_sanity(self):
        parser = PythonLanguageParser()

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'print "Hello world"')
        self.assertEquals(parseresult.keywords['print'], 1)
        for keyword in parser.keywords:
            if keyword != 'print':
                self.assertEquals(parseresult.keywords[keyword], 0)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'if x == 10: pass')
        self.assertEquals(parseresult.keywords['if'], 1)
        self.assertEquals(parseresult.keywords['pass'], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'while True: pass')
        self.assertEquals(parseresult.keywords['while'], 1)
        self.assertEquals(parseresult.keywords['pass'], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'for x in [1, 2]: pass')
        self.assertEquals(parseresult.keywords['for'], 1)
        self.assertEquals(parseresult.keywords['in'], 1)
        self.assertEquals(parseresult.keywords['pass'], 1)

    def test_operators_sanity(self):
        parser = PythonLanguageParser()

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a == 10')
        self.assertEquals(parseresult.operators['=='], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a==10')
        self.assertEquals(parseresult.operators['=='], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a*10')
        self.assertEquals(parseresult.operators['*'], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a ==10')
        self.assertEquals(parseresult.operators['=='], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a== 10')
        self.assertEquals(parseresult.operators['=='], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a > 10')
        self.assertEquals(parseresult.operators['>'], 1)

        parseresult = parser.make_parseresult()
        parser.parse(parseresult, 'a <> 10')
        self.assertEquals(parseresult.operators['<>'], 1)



import unittest
from detektor.functionextractor.clike import ClikeFunctionExtractor


class TestClikeFunctionExtractor(unittest.TestCase):

    def test_extract_function(self):
        keywords = {
            'if', 'while'
        }
        functions = ClikeFunctionExtractor(keywords).extract("""
            void main(int i) {
                // Some code here
            }

            int[] test() {
            }

            Array<String> anotherTest() {
                if(i == 10) {
                    while(true) {
                        print("Something");
                    }
                }
            }
        """)
        self.assertEquals(len(functions), 3)
        self.assertEquals(functions[0].name, 'void main(int i)')
        self.assertEquals(functions[1].name, 'int[] test()')
        self.assertEquals(functions[2].name, 'Array<String> anotherTest()')
        self.assertEquals(functions[0].sourcecode.strip(), '// Some code here')




class ParseResult(object):
    """
    Parse result abstract class.

    The parsers produce subclasses of :class:`.EditableParseResult`. That is OK for
    small datasets, but when you compare huge amounts of code you will want to
    serialize the results to avoid re-processing the same peace of code. To do
    this, you use meth:`EditableParseResult.serialize_as_dict`, and
    load it later with :meth:`.UneditableParseResult`.

    All code for comparing parse results should use the methods in this interface,
    not the attributes made available by EditableParseResult or UneditableParseResult.
    """
    def __unicode__(self):
        return \
            u'{classname}('\
            u'keywords={keywords!r}, '\
            u'operators={operators!r})'.format(
                classname=self.__class__.__name__,
                keywords=self.get_keywords_string(),
                operators=self.get_operators_string())

    def __str__(self):
        return unicode(self).encode('ascii', 'replace')

    def __repr__(self):
        return str(self)

    def get_codeblocktype(self):
        raise NotImplementedError()

    def get_label(self):
        raise NotImplementedError()

    def get_operators_string(self):
        raise NotImplementedError()

    def get_keywords_string(self):
        raise NotImplementedError()

    def get_number_of_operators(self):
        raise NotImplementedError()

    def get_number_of_keywords(self):
        raise NotImplementedError()

    def get_operators_and_keywords_string(self):
        raise NotImplementedError()

    def get_normalized_sourcecode(self):
        raise NotImplementedError()

    def get_parsed_functions(self):
        raise NotImplementedError()


class UneditableParseResult(object):
    """

    """
    def __init__(self, dumped_dict):
        """
        Parameters:
            dumped_dict: A dict produced by
        """
        self._dumped_dict = dumped_dict
        self._parsed_functions = None

    def get_codeblocktype(self):
        return self._dumped_dict['codeblocktype']

    def get_label(self):
        return self._dumped_dict['label']

    def get_operators_string(self):
        return self._dumped_dict['operators_string']

    def get_keywords_string(self):
        return self._dumped_dict['keywords_string']

    def get_number_of_operators(self):
        return self._dumped_dict['number_of_operators']

    def get_number_of_keywords(self):
        return self._dumped_dict['number_of_keywords']

    def get_operators_and_keywords_string(self):
        return self._dumped_dict['operators_and_keywords_string']

    def get_normalized_sourcecode(self):
        return self._dumped_dict['normalized_sourcecode']

    def get_parsed_functions(self):
        if self._parsed_functions is None:
            self._parsed_functions = map(UneditableParseResult, self._dumped_dict['parsed_functions'])


class EditableParseResult(ParseResult):
    """
    Language parsers create objects of this class to store their results.
    """

    def __init__(self, codeblocktype, label, set_of_all_operators, set_of_all_keywords):
        """
        Parameters:
            codeblocktype: The type of code this result is for. Must be one of: "program", "function".
            label: A label that can be set to give the ParseResult a name/context.
                Used to name functions in the list of functions
                when the codeblocktype is ``function``.
            set_of_all_keywords (set): A set of keywords supported by this language.
                used to create stable digest of keywords and the number of occurrences.
            set_of_all_operators (set): A set of operators supported by this language.
                used to create stable digest of operators and the number of occurrences.
        """
        self.codeblocktype = codeblocktype
        self.label = label

        # Dict mapping operators to the number of occurrences of that operator.
        self.operators = dict.fromkeys(set_of_all_operators, 0)

        # Dict mapping keywords to the number of occurrences of that keyword.
        self.keywords = dict.fromkeys(set_of_all_keywords, 0)

        self.number_of_operators = 0
        self.number_of_keywords = 0

        # A string with all the operators and keywords appended in the order
        # they occur in the code. Handled automatically in :meth:`.add_keyword`
        # and :meth:`.add_operator`, so you should not have to edit this.
        self.operators_and_keywords_string = u''

        # Should be a string that when compared with another ParseResult
        # means that the ParseResults are very similar if they match.
        # This is typically populated by a smart language parser that
        # normalizes sourcecode in a way that makes it very unlikely that
        # programs differ if this string is 100% equal in two programs.
        # If the language does not use this, it should be left as ``None``,
        # to indicate that this should be ignored when comparing programs.
        self.normalized_sourcecode = None

        # List of class:`.ParseResult` objects - one for each parsed function.
        # This is only used when ``codeblocktype`` is ``program``.
        self.parsed_functions = []

    def is_program(self):
        return self.codeblocktype == 'program'

    def add_keyword(self, keyword):
        self.number_of_keywords += 1
        self.keywords[keyword] += 1
        self.operators_and_keywords_string += keyword

    def add_operator(self, operator):
        self.number_of_operators += 1
        self.operators[operator] += 1
        self.operators_and_keywords_string += operator

    def _make_string_from_count_dict(self, dct):
        return '_'.join([str(dct[key]) for key in sorted(dct.keys())])

    def get_codeblocktype(self):
        return self.codeblocktype

    def get_label(self):
        return self.label

    def get_operators_string(self):
        return self._make_string_from_count_dict(self.operators)

    def get_keywords_string(self):
        return self._make_string_from_count_dict(self.keywords)

    def get_number_of_operators(self):
        return self.number_of_operators

    def get_number_of_keywords(self):
        return self.number_of_keywords

    def get_operators_and_keywords_string(self):
        return self.operators_and_keywords_string

    def get_normalized_sourcecode(self):
        return self.normalized_sourcecode

    def get_parsed_functions(self):
        return self.parsed_functions

    def serialize_as_dict(self):
        return {
            'codeblocktype': self.get_codeblocktype(),
            'label': self.get_label(),
            'operators_string': self.get_operators_string(),
            'keywords_string': self.get_keywords_string(),
            'number_of_operators': self.get_number_of_operators(),
            'number_of_keywords': self.get_number_of_keywords(),
            'operators_and_keywords_string': self.get_operators_and_keywords_string(),
            'normalized_sourcecode': self.get_normalized_sourcecode(),
            'parsed_functions': [function.serialize_as_dict() for function in self.parsed_functions],
        }



from detektor.languageparser.java import JavaLanguageParser
from detektor.languageparser.python import PythonLanguageParser


default_languageparsers = {
    'java': JavaLanguageParser,
    'python': PythonLanguageParser,
}

def make_parser(language):
    """
    Make a Detektor language parser for the given language.

    Example::

        parser = make_parser('python')
        sourcecodes = [
            'print "hello world"',
            'print "This is a test"'
        ]

        # Treat both the sourcecodes as a single "program",
        # and collect the results in a single ParseResult object:
        parseresult = parser.make_parseresult()
        for sourcecode in sourcecodes:
            parser.parse(sourcecode, parseresult)
        print parseresult

        # ... or treat them as separate programs:
        results = []
        for sourcecode in sourcecodes:
            parseresult = parser.make_parseresult()
            parser.parse(sourcecode, parseresult)
            results.append(parseresult)

    """
    return default_languageparsers[language]()



from detektor.codefilter.normalizeoperatorwhitespace import NormalizeOperatorWhitespace
from detektor.languageparser.shlexparser import ShlexLanguageParserBase
from detektor.functionextractor.clike import ClikeFunctionExtractor


class JavaLanguageParser(ShlexLanguageParserBase):
    # http://docs.oracle.com/javase/tutorial/java/nutsandbolts/opsummary.html
    operators = {
        '!', '+', '-', '*', '**', '/', '//', '%', '<<', '>>', '>>>', '&', '|',
        '++', '--', '&&', '||', '?:',
        '=', '+=', '-=', '*=',
        '^', '~', '<', '>', '<=', '>=', '==', '!=',
    }

    # http://docs.oracle.com/javase/tutorial/java/nutsandbolts/_keywords.html
    keywords = {
        'abstract', 'continue', 'for', 'new', 'switch',
        'assert', 'default', 'package', 'synchronized',
        'boolean', 'do', 'if', 'private', 'this',
        'break', 'double', 'implements', 'protected', 'throw',
        'byte', 'else', 'import', 'public', 'throws',
        'case', 'enum', 'instanceof', 'return', 'transient',
        'catch', 'extends', 'int', 'short', 'try',
        'char', 'final', 'interface', 'static', 'void',
        'class', 'finally', 'long', 'strictfp', 'volatile',
        'float', 'native', 'super', 'while',
    }

    sourcecode_preprocessor_classes = [
        NormalizeOperatorWhitespace
    ]

    def __init__(self, *args, **kwargs):
        super(JavaLanguageParser, self).__init__(*args, **kwargs)
        self.function_extractor = ClikeFunctionExtractor(keywords=self.keywords)

    def extract_functionsourcecode(self, sourcecode):
        return self.function_extractor.extract(sourcecode)



class CompareTwo(object):
    matchmap = {
        'operators_and_keywords_string_equal': {
            'points': 10,
            'label': 'Programs very similar. Look for use of query-replace.'
        },
        'operators_string_equal': {
            'points': 3,
            'label': 'Equal number of each operator used.'
        },
        'keywords_string_equal': {
            'points': 3,
            'label': 'Equal number of each keyword used.'
        },
        'total_operatorcount_equal': {
            'points': 1,
            'label': 'Equal number of total operators.'
        },
        'total_keywordcount_equal': {
            'points': 1,
            'label': 'Equal number of total keywords.'
        },
        'similar_functions': {
            # NOTE: No points for similar functions, they are collected
            # for each similarity, and scaled according to
            'label': 'Some of the functions look similar.'
        }
    }

    def __init__(self, parseresult1, parseresult2, scale=100, functionscale=1):
        """
        Parameters:
            parseresult1: A obj:`detektor.parseresult.ParseResults` object.
            parseresult2: A obj:`detektor.parseresult.ParseResults` object.
            scale (int): A number to multiply all the points produced on
                matches.
            functionscale (int): A number to multiply all the points produced on
                matches in functions. Should normally be a fair bit smaller
                than ``scale``.
        """
        self.parseresult1 = parseresult1
        self.parseresult2 = parseresult2
        self.scale = scale
        self.functionscale = functionscale

        # Points are collected by the _compare() function.
        # When the comparison has completed, the number of points indicates
        # how similar the two programs are.
        # You normally get this via meth:`.get_scaled_points`.
        self.points = 0

        # Just like ``self.points``, but collected when comparing functions.
        # You normally get this via meth:`.get_scaled_points`.
        self.functionpoints = 0

        #: List of matchids. We use get_description_for_matchid() to get a human
        #: readable string for a matchid.
        self.summary = []

        #: List of :obj:`.CompareTwo` objects for all
        #: functions within the two ParseResult objects.
        self.comparetwo_for_functions = []

    def __unicode__(self):
        return u'{parseresult1},{parseresult2} - Points: {points}, summary: {summary}'.format(
            parseresult1=self.parseresult1.get_label(),
            parseresult2=self.parseresult2.get_label(),
            points=self.get_scaled_points(),
            summary=self.get_summary_descriptions_as_string())

    def __str__(self):
        return unicode(self).encode('ascii', 'replace')

    def __repr__(self):
        return str(self)

    def get_scaled_points(self):
        return self.points * self.scale + self.functionpoints * self.functionscale

    def get_summary_descriptions_as_list(self):
        return [self.get_description_for_matchid(matchid) for matchid in self.summary]

    def get_summary_descriptions_as_string(self):
        return u' '.join(self.get_summary_descriptions_as_list())

    def get_description_for_matchid(self, matchid):
        return self.matchmap[matchid]['label']

    def get_points_for_matchid(self, matchid):
        return self.matchmap[matchid]['points']

    def _add_match_if_matched(self, matchid):
        if matchid is not None:
            self.summary.append(matchid)
            self.points += self.matchmap[matchid].get('points', 0)

    def compare(self):
        self._add_match_if_matched(self._compare_operators_and_keywords_string_equal())
        self._add_match_if_matched(self._compare_operators_string_equal())
        self._add_match_if_matched(self._compare_keywords_string_equal())
        self._add_match_if_matched(self._compare_total_operatorcount_equal())
        self._add_match_if_matched(self._compare_total_keywordcount_equal())
        if self.parseresult1.get_codeblocktype() == 'program' and self.parseresult2.get_codeblocktype() == 'program':
            self._add_match_if_matched(self._compare_functions())

    def compares_parseresults(self, parseresulta, parseresultb):
        """
        Returns True if the given ParseResult objects are the ones beeing compared
        by this CompareTwo. Order does not matter.
        """
        return {parseresulta, parseresultb} == {self.parseresult1, self.parseresult2}

    #
    #
    # Compare-methods
    # ===============
    # Each of them return a key from :obj:`.matchmap` if they find a match.
    # If they do not return None, we use :meth:`._add_match_if_matched` to
    # look them up in :obj:`.matchmap` to find the points awarded for the
    # match.
    #
    #

    def _compare_operators_and_keywords_string_equal(self):
        a = self.parseresult1
        b = self.parseresult2
        if a.get_operators_and_keywords_string() == b.get_operators_and_keywords_string():
            return 'operators_and_keywords_string_equal'
        else:
            return None

    def _compare_operators_string_equal(self):
        if self.parseresult1.get_operators_string() == self.parseresult2.get_operators_string():
            return 'operators_string_equal'
        else:
            return None

    def _compare_keywords_string_equal(self):
        if self.parseresult1.get_keywords_string() == self.parseresult2.get_keywords_string():
            return 'keywords_string_equal'
        else:
            return None

    def _compare_total_operatorcount_equal(self):
        if self.parseresult1.get_number_of_operators() == self.parseresult2.get_number_of_operators():
            return 'total_operatorcount_equal'
        else:
            return None

    def _compare_total_keywordcount_equal(self):
        if self.parseresult1.get_number_of_keywords() == self.parseresult2.get_number_of_keywords():
            return 'total_keywordcount_equal'
        else:
            return None

    def _compare_functions(self):
        self.comparetwo_for_functions = []
        functionparseresults = self.parseresult1.get_parsed_functions()
        for functionparseresult1 in self.parseresult1.get_parsed_functions():
            for functionparseresult2 in self.parseresult2.get_parsed_functions():
                comparetwo = CompareTwo(
                    functionparseresult1, functionparseresult2,
                    scale=self.functionscale)
                comparetwo.compare()
                self.comparetwo_for_functions.append(comparetwo)
                self.functionpoints += comparetwo.points
        if self.functionpoints:
            return 'similar_functions'
        else:
            return None


class CompareMany(object):
    """
    Used to compare many :class:`detektor.parseresult.ParseResult` objects.
    """
    def __init__(self, parseresults):
        """
        Parameters:
            parseresults (iterable):
                List/iterable of :class:`detektor.parseresult.ParseResult` objects.
        """
        self.results = []
        for index1, parseresult1 in enumerate(parseresults):
            for index2 in xrange(index1 + 1, len(parseresults)):
                parseresult2 = parseresults[index2]
                self.compare(parseresult1, parseresult2)

    def sort_by_points_descending(self):
        """
        Sort by points in place.
        """
        self.results.sort(cmp=lambda a, b: cmp(b.points, a.points))

    def __iter__(self):
        return self.results.__iter__()

    def compare(self, parseresult1, parseresult2):
        """
        Called once for each combination of all the ParseResult objects.
        Can be overridden to tune the comparison code.
        """
        comparetwo = CompareTwo(parseresult1, parseresult2)
        comparetwo.compare()
        self.add_to_results(comparetwo)

    def add_to_results(self, comparetwo):
        """
        Called once after each comparison of two parseresults are
        completed.

        Defaults to adding the given :class:`.CompareTwo`
        object to :obj:`.results`, but you can override this to
        limit what to add to results.

        Typically you may want to override this to avoid adding
        results with zero points::

            class MyCompareMany(detektor.comparer.CompareMany):
                def add_to_results(self, comparetwo):
                    if comparetwo.get_scaled_points() > 0:
                        super(MyCompareMany, self).add_to_results(comparetwo)
        """
        self.results.append(comparetwo)





class CodeFilterBase(object):
    """
    A codefilter is a peace of code that does something to some blob of
    sourcecode and returns the modified sourcecode.

    It is perfect for making reusable filters that normalized code
    as part of parsing the code.

    Note that functionally, you can just as easily override the
    :meth:`~detektor.languageparser.base.LanguageParserBase.parse` method
    of a parser and do this preprocessing, but that makes it hard-coupled
    to that parser (which is bad for filters/preprocessors that many
    languages may need).
    """

    def __init__(self, languageparser):
        """
        Parameters:
            languageparser: The :class:`detektor.languageparser.base.LanguageParserBase`
                this filter is to be applied to.
        """
        self.languageparser = languageparser

    def filter(self, sourcecode):
        """
        This method performs the filtering.
        Must be overridden in subclasses.

        Parameters:
            sourcecode: The sourcecode to filter.

        Returns:
            The modified ``sourcecode``.
        """
        raise NotImplementedError()



from setuptools import setup, find_packages

setup(
    name='detektor',
    description='A library for finding similarities in code.',
    version='1.1.0-beta.005',
    url='https://github.com/appressoas/detektor',
    author='Magne Westlie, Espen Angell Kristiansen, Tor Ivar Johansen',
    author_email='magne@appresso.no,espen@appresso.no,tor@appresso.no',
    license='GPL',
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    install_requires=['setuptools'],
    test_suite='detektor.tests',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ]
)



import unittest
import mock
from detektor.comparer import CompareTwo, CompareMany


class TestCompareTwo(unittest.TestCase):

    def test_compare_operators_and_keywords_string_equal_nomatch(self):
        a = mock.MagicMock()
        a.get_operators_and_keywords_string.return_value = 'a'
        b = mock.MagicMock()
        b.get_operators_and_keywords_string.return_value = 'b'
        self.assertEqual(
            CompareTwo(a, b)._compare_operators_and_keywords_string_equal(),
            None)

    def test_compare_operators_and_keywords_string_equal(self):
        a = mock.MagicMock()
        a.get_operators_and_keywords_string.return_value = 'x'
        b = mock.MagicMock()
        b.get_operators_and_keywords_string.return_value = 'x'
        self.assertEqual(
            CompareTwo(a, b)._compare_operators_and_keywords_string_equal(),
            'operators_and_keywords_string_equal')

    def test_compare_operators_string_equal_nomatch(self):
        a = mock.MagicMock()
        a.get_operators_string.return_value = 'a'
        b = mock.MagicMock()
        b.get_operators_string.return_value = 'b'
        self.assertEqual(
            CompareTwo(a, b)._compare_operators_string_equal(),
            None)

    def test_compare_operators_string_equal(self):
        a = mock.MagicMock()
        a.get_operators_string.return_value = 'x'
        b = mock.MagicMock()
        b.get_operators_string.return_value = 'x'
        self.assertEqual(
            CompareTwo(a, b)._compare_operators_string_equal(),
            'operators_string_equal')

    def test_compare_keywords_string_equal_nomatch(self):
        a = mock.MagicMock()
        a.get_keywords_string.return_value = 'a'
        b = mock.MagicMock()
        b.get_keywords_string.return_value = 'b'
        self.assertEqual(
            CompareTwo(a, b)._compare_keywords_string_equal(),
            None)

    def test_compare_keywords_string_equal(self):
        a = mock.MagicMock()
        a.get_keywords_string.return_value = 'x'
        b = mock.MagicMock()
        b.get_keywords_string.return_value = 'x'
        self.assertEqual(
            CompareTwo(a, b)._compare_keywords_string_equal(),
            'keywords_string_equal')

    def test_compare_total_operatorcount_equal_nomatch(self):
        a = mock.MagicMock()
        a.get_number_of_operators.return_value = 10
        b = mock.MagicMock()
        b.get_number_of_operators.return_value = 20
        self.assertEqual(
            CompareTwo(a, b)._compare_total_operatorcount_equal(),
            None)

    def test_compare_total_operatorcount_equal(self):
        a = mock.MagicMock()
        a.get_number_of_operators.return_value = 5
        b = mock.MagicMock()
        b.get_number_of_operators.return_value = 5
        self.assertEqual(
            CompareTwo(a, b)._compare_total_operatorcount_equal(),
            'total_operatorcount_equal')

    def test_compare_total_keywordcount_equal_nomatch(self):
        a = mock.MagicMock()
        a.get_number_of_keywords.return_value = 10
        b = mock.MagicMock()
        b.get_number_of_keywords.return_value = 20
        self.assertEqual(
            CompareTwo(a, b)._compare_total_keywordcount_equal(),
            None)

    def test_compare_total_keywordcount_equal(self):
        a = mock.MagicMock()
        a.get_number_of_keywords.return_value = 5
        b = mock.MagicMock()
        b.get_number_of_keywords.return_value = 5
        self.assertEqual(
            CompareTwo(a, b)._compare_total_keywordcount_equal(),
            'total_keywordcount_equal')

    def test_compare_functions(self):
        parsedfunction1 = mock.MagicMock()
        parsedfunction1.get_number_of_keywords.return_value = 5
        parsedfunction2 = mock.MagicMock()
        parsedfunction2.get_number_of_keywords.return_value = 2
        parsedfunction3 = mock.MagicMock()
        parsedfunction3.get_number_of_keywords.return_value = 3
        parsedfunction4 = mock.MagicMock()
        parsedfunction4.get_number_of_keywords.return_value = 5

        a = mock.MagicMock()
        a.codeblocktype = 'program'
        a.get_parsed_functions.return_value = [parsedfunction1, parsedfunction2]
        b = mock.MagicMock()
        b.codeblocktype = 'program'
        b.get_parsed_functions.return_value = [parsedfunction3, parsedfunction4]
        comparetwo = CompareTwo(a, b)
        functioncompare_summary = comparetwo._compare_functions()
        self.assertEqual(functioncompare_summary, 'similar_functions')
        self.assertEqual(comparetwo.functionpoints, 1)
        self.assertEqual(comparetwo.points, 0)
        self.assertEqual(comparetwo.get_scaled_points(), 1)

    def test_compare_functions_complex(self):
        parsedfunctionA1 = mock.MagicMock()
        parsedfunctionA1.get_number_of_keywords.return_value = 5  # 1 point
        parsedfunctionA1.get_operators_and_keywords_string.return_value = 'ifreturn'  # 10 points
        parsedfunctionA2 = mock.MagicMock()
        parsedfunctionA2.get_number_of_keywords.return_value = 2
        parsedfunctionA2.get_operators_string.return_value = 'x'  # 3 points

        parsedfunctionB1 = mock.MagicMock()
        parsedfunctionB1.get_number_of_keywords.return_value = 3
        parsedfunctionB1.get_operators_and_keywords_string.return_value = 'ifreturn'  # 10 points
        parsedfunctionB2 = mock.MagicMock()
        parsedfunctionB2.get_number_of_keywords.return_value = 5  # 1 point
        parsedfunctionB2.get_operators_string.return_value = 'x'  # 3 points

        a = mock.MagicMock()
        a.codeblocktype = 'program'
        a.get_parsed_functions.return_value = [parsedfunctionA1, parsedfunctionA2]
        b = mock.MagicMock()
        b.codeblocktype = 'program'
        b.get_parsed_functions.return_value = [parsedfunctionB1, parsedfunctionB2]
        comparetwo = CompareTwo(a, b)
        functioncompare_summary = comparetwo._compare_functions()
        self.assertEqual(functioncompare_summary, 'similar_functions')
        self.assertEqual(comparetwo.functionpoints, 14)


class TestParseResultCompareMany(unittest.TestCase):
    def test_all_compared_single(self):
        parseresult1 = mock.MagicMock()
        comparemany = CompareMany([parseresult1])
        self.assertEquals(comparemany.results, [])

    def test_all_compared_two(self):
        parseresult1 = mock.MagicMock()
        parseresult2 = mock.MagicMock()
        comparemany = CompareMany([parseresult1, parseresult2])
        self.assertEquals(len(comparemany.results), 1)
        self.assertEquals(comparemany.results[0].parseresult1, parseresult1)
        self.assertEquals(comparemany.results[0].parseresult2, parseresult2)

    def test_all_compared_three(self):
        parseresult1 = mock.MagicMock()
        parseresult2 = mock.MagicMock()
        parseresult3 = mock.MagicMock()
        comparemany = CompareMany([parseresult1, parseresult2, parseresult3])
        results = comparemany.results
        self.assertEquals(len(results), 3)
        self.assertTrue(results[0].compares_parseresults(parseresult1, parseresult2))
        self.assertTrue(results[1].compares_parseresults(parseresult1, parseresult3))
        self.assertTrue(results[2].compares_parseresults(parseresult2, parseresult3))

    def test_all_compared_four(self):
        parseresult1 = mock.MagicMock()
        parseresult2 = mock.MagicMock()
        parseresult3 = mock.MagicMock()
        parseresult4 = mock.MagicMock()
        comparemany = CompareMany([parseresult1, parseresult2, parseresult3, parseresult4])
        results = comparemany.results
        self.assertEquals(len(results), 6)
        self.assertTrue(results[0].compares_parseresults(parseresult1, parseresult2))
        self.assertTrue(results[1].compares_parseresults(parseresult1, parseresult3))
        self.assertTrue(results[2].compares_parseresults(parseresult1, parseresult4))
        self.assertTrue(results[3].compares_parseresults(parseresult2, parseresult3))
        self.assertTrue(results[4].compares_parseresults(parseresult2, parseresult4))
        self.assertTrue(results[5].compares_parseresults(parseresult3, parseresult4))

    def test_compare_many_results_sanity(self):
        parseresult1 = mock.MagicMock()
        parseresult1.label = '1'
        parseresult1.get_operators_and_keywords_string.return_value = 'x'

        parseresult2 = mock.MagicMock()
        parseresult2.label = '2'
        parseresult2.get_operators_and_keywords_string.return_value = 'x'
        parseresult2.get_number_of_operators.return_value = 10

        parseresult3 = mock.MagicMock()
        parseresult3.label = '3'
        parseresult3.get_number_of_operators.return_value = 10

        comparemany = CompareMany([parseresult1, parseresult2, parseresult3])
        # for comparetwo in comparemany:
        #     print comparetwo
        results = comparemany.results
        self.assertTrue(results[0].compares_parseresults(parseresult1, parseresult2))
        self.assertEquals(results[0].points, 10)
        self.assertEquals(results[0].get_scaled_points(), 1000)
        self.assertEquals(results[0].summary, ['operators_and_keywords_string_equal'])
        self.assertTrue(results[1].compares_parseresults(parseresult1, parseresult3))
        self.assertEquals(results[1].points, 0)
        self.assertEquals(results[1].get_scaled_points(), 0)
        self.assertEquals(results[1].summary, [])
        self.assertTrue(results[2].compares_parseresults(parseresult2, parseresult3))
        self.assertEquals(results[2].points, 1)
        self.assertEquals(results[2].get_scaled_points(), 100)
        self.assertEquals(results[2].summary, ['total_operatorcount_equal'])
